package com.github.snowdream.android.widget;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Bitmap.CompressFormat;
import android.graphics.BitmapFactory;
import android.text.TextUtils;
import android.webkit.URLUtil;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.SoftReference;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class SmartImageCache {
    private static final String DISK_CACHE_PATH = "/web_image_cache/";

    private ConcurrentHashMap<String, SoftReference<Bitmap>> memoryCache;
    private static String diskCachePath = null;
    private static boolean diskCacheEnabled = false;
    private ExecutorService writeThread;
    private ConcurrentHashMap<String,Object> mCancelList = null;

    public SmartImageCache(Context context) {
        // Set up in-memory cache store
        memoryCache = new ConcurrentHashMap<String, SoftReference<Bitmap>>();

        //set up the cancel list
        mCancelList = new ConcurrentHashMap<String,Object>();
        
        if (TextUtils.isEmpty(diskCachePath)) {
            // Set up disk cache store
            Context appContext = context.getApplicationContext();
            diskCachePath = appContext.getCacheDir().getAbsolutePath() + DISK_CACHE_PATH;

            File outFile = new File(diskCachePath);
            outFile.mkdirs();

            diskCacheEnabled = outFile.exists();
        }

        // Set up threadpool for image fetching tasks
        writeThread = Executors.newSingleThreadExecutor();
        //writeThread = ExecutorDelegate.getExecutorService();

    }

    public static void setImageCache(String cacheDir){
        diskCachePath = cacheDir;

        File outFile = new File(diskCachePath);
        outFile.mkdirs();

        diskCacheEnabled = outFile.exists();
    }

    public void cancel(final String url){
        mCancelList.put(url,"");
    }
    
    public boolean isEnabled(final String url){
        boolean ret = true;
        
        if (mCancelList != null && mCancelList.containsKey(url)) {
            ret = false;
        }
    
        return ret;
    }
    
    public Bitmap get(final String url) {
        Bitmap bitmap = null;

        if (URLUtil.isNetworkUrl(url)){
            // Check for image in memory
            bitmap = getBitmapFromMemory(url);

            // Check for image on disk cache
            if(bitmap == null) {
                bitmap = getBitmapFromDisk(url);

                // Write bitmap back into memory cache
                if(bitmap != null) {
                    cacheBitmapToMemory(url, bitmap);
                }
            }
        }else {
            // Check for image in memory
            bitmap = getBitmapFromMemory(url);
        }

        return bitmap;
    }

    public Bitmap get(final String url,int width, int height) {
        Bitmap bitmap = null;

        if (URLUtil.isNetworkUrl(url)){
            // Check for image in memory
            bitmap = getBitmapFromMemory(url);

            // Check for image on disk cache
            if(bitmap == null) {
                bitmap = getBitmapFromDisk(url, width, height);

                // Write bitmap back into memory cache
                if(bitmap != null) {
                    cacheBitmapToMemory(url, bitmap);
                }
            }
        }else {
            // Check for image in memory
            bitmap = getBitmapFromMemory(url);
        }

        return bitmap;
    }

    public void put(String url, Bitmap bitmap) {
        //put into memory cache
        cacheBitmapToMemory(url, bitmap);

        //put into file cache
        if (URLUtil.isNetworkUrl(url)){
            cacheBitmapToDisk(url, bitmap);
        }
    }

    public void put(String url,InputStream is, int width, int height){
        //put into file cache
        if (URLUtil.isNetworkUrl(url)){
            cacheInputStreamToDisk(url, is, width, height);
        }
    }


    public void remove(String url) {
        if(url == null){
            return;
        }

        // Remove from memory cache
        memoryCache.remove(getCacheKey(url));

        if (URLUtil.isNetworkUrl(url)){
            // Remove from file cache
            String filePath = getFilePath(url);
            File f = new File(filePath);
            if(f.exists() && f.isFile()) {
                f.delete();
            }
        }
    }

    public void clear() {
        // Remove everything from memory cache
        memoryCache.clear();

        // Remove everything from file cache
        File cachedFileDir = new File(diskCachePath);
        if(cachedFileDir.exists() && cachedFileDir.isDirectory()) {
            File[] cachedFiles = cachedFileDir.listFiles();
            for(File f : cachedFiles) {
                if(f.exists() && f.isFile()) {
                    f.delete();
                }
            }
        }
    }

    private void cacheBitmapToMemory(final String url, final Bitmap bitmap) {
        memoryCache.put(getCacheKey(url), new SoftReference<Bitmap>(bitmap));
    }

    private void cacheBitmapToDisk(final String url, final Bitmap bitmap) {
        writeThread.execute(new Runnable() {
            @Override
            public void run() {
                File file = new File(diskCachePath, getCacheKey(url));
                if(isEnabled(url) && diskCacheEnabled) {
                    BufferedOutputStream ostream = null;
                    try {
                        ostream = new BufferedOutputStream(new FileOutputStream(file), 2*1024);
                        bitmap.compress(CompressFormat.PNG, 100, ostream);
                    } catch (FileNotFoundException e) {
                        e.printStackTrace();
                    } finally {
                        try {
                            if(ostream != null) {
                                ostream.flush();
                                ostream.close();
                            }
                        } catch (IOException e) {}
                    }
                }
                
                if (!isEnabled(url)) {
                    file.delete();
                    mCancelList.remove(url);
                }
            }
        });
    }

    private void cacheInputStreamToDisk(final String url, final InputStream is, int width, int height) {
        if(diskCacheEnabled) {
            File file = new File(diskCachePath, getCacheKey(url));

            BufferedOutputStream fos = null;
            try {
                fos = new BufferedOutputStream(new FileOutputStream(file), 2*1024);

                // 将输入流is写入文件输出流fos中  
                int ch = 0;  
                while(isEnabled(url) &&  (ch=is.read()) != -1){  
                    fos.write(ch);  
                }  
                
                if (!isEnabled(url)) {
                    file.delete();
                    mCancelList.remove(url);
                }

            }catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } finally {
                if(fos != null) {
                    try {
                        fos.flush();
                        fos.close();
                    } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }

                if (is != null) {
                    try {
                        is.close();
                    } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }

            if (file.exists()) {
                cacheDiskToMemory(url, file, width, height);
            }
        }
    }

    private void cacheDiskToMemory(final String url, final File file, int width, int height) {
        Bitmap bitmap = BitmapUtils.getBitmapFromFile(file,width,height);
        if (bitmap != null) {
            memoryCache.put(getCacheKey(url), new SoftReference<Bitmap>(bitmap));
        }
    }

    private Bitmap getBitmapFromMemory(String url) {
        Bitmap bitmap = null;
        SoftReference<Bitmap> softRef = memoryCache.get(getCacheKey(url));
        if(softRef != null){
            bitmap = softRef.get();
        }

        return bitmap;
    }

    private Bitmap getBitmapFromDisk(String url) {
        Bitmap bitmap = null;
        if(diskCacheEnabled){
            String filePath = getFilePath(url);
            File file = new File(filePath);
            if(file.exists()) {
                bitmap = BitmapFactory.decodeFile(filePath);
            }
        }
        return bitmap;
    }

    private Bitmap getBitmapFromDisk(String url, int width, int height) {
        Bitmap bitmap = null;

        File file = new File(diskCachePath, getCacheKey(url));
        if (file.exists()) {
            bitmap = BitmapUtils.getBitmapFromFile(file,width,height);
        }

        return bitmap;
    }    


    private String getFilePath(String url) {
        return diskCachePath + File.separator + getCacheKey(url);
    }

    private String getCacheKey(String url) {
        if(url == null){
            throw new RuntimeException("Null url passed in");
        } else {
            return url.replaceAll("[.:/,%?&=]", "+").replaceAll("[+]+", "+");
        }
    }
}